@@ -21,6 +21,7 @@ from thunar_plugins.version import __version__
 
 # external modules
 from rich.pretty import Pretty
+from rich.syntax import Syntax
 import gi
 
 gi.require_version("Gtk", "3.0")
@@ -34,6 +35,45 @@ METADATA_CACHE = defaultdict(lambda: defaultdict(set))
 
 
 class GitAnnexHelper:
+    @functools.lru_cache(maxsize=1)
+    def git_annex_executable(self):
+        return shutil.which("git-annex")
+
+    @functools.lru_cache(maxsize=1)
+    def git_annex_version(self):
+        version = self.subprocess(
+            subprocess.check_output, ["git", "annex", "version", "--raw"]
+        )
+        logger.info(f"Git Annex version: {version}")
+        return version
+
+    @functools.lru_cache(maxsize=1)
+    def git_annex_capabilities(self):
+        caps = set()
+        view_help = self.subprocess(
+            subprocess.check_output, ["git", "annex", "view", "--help"]
+        )
+        if re.search(r"FIELD\?=", view_help, flags=re.IGNORECASE):
+            caps.add("view-show-valueless")
+        logger.info(
+            f"This git-annex ({self.git_annex_version()}) "
+            f"{'can' if 'view-show-valueless' in caps else 'can´t'} "
+            f"show unvalued fields in a view."
+        )
+        logger.debug(f"git annex capabilities: {caps}")
+        return caps
+
+    @classmethod
+    def debug_cmdline(cls, cmdparts, comment="📋 Copy-pastable:"):
+        if logger.getEffectiveLevel() <= logging.DEBUG:
+            console.print(
+                Syntax(
+                    f"\n# {comment}\n\n{shlex.join(cmdparts)}\n",
+                    "bash",
+                    word_wrap=True,
+                )
+            )
+
     @classmethod
     def flush_gtk(cls):
         while Gtk.events_pending():
@@ -122,19 +162,23 @@ class GitAnnexHelper:
             cache = defaultdict(lambda: defaultdict(set))
         notification.update(
             _("Git Annex Thunar Plugin"),
-            "⏳ " + _("Determining amount of files..."),
+            "⏳ "
+            + _("Determining amount of annexed files in {repo}").format(
+                repo=pathlib.Path(path).name
+            ),
             "git-annex",
         )
         notification.show()
+        n_files = 0
         for n_files, f in enumerate(
             cls.subprocess(
                 subprocess.Popen,
-                ["git", "-C", path, "ls-files"],
+                ["git", "-C", path, "annex", "find", "--copies=0"],
                 stdout=subprocess.PIPE,
             ).stdout,
             start=1,
         ):
-            pass
+            cls.flush_gtk()
         try:
             git_annex = cls.subprocess(
                 subprocess.Popen,
@@ -145,7 +189,9 @@ class GitAnnexHelper:
 
             time_last_notification_update = 0
             time_begin = time.time()
+            i = 0
             for i, line in enumerate(git_annex.stdout, start=1):
+                cls.flush_gtk()
                 if time.time() - time_last_notification_update > 1:
                     # couldn't get a stop button to work, callback was never
                     # called, don't know why 🤷
@@ -159,11 +205,7 @@ class GitAnnexHelper:
                         )
                         + (
                             "\n\n{}".format(
-                                _(
-                                    "Sorry, you can't interrupt this "
-                                    "and Thunar is frozen until it's finished."
-                                )
-                                + "🙂 "
+                                _("This currently can't be interrupted") + "🙂 "
                             )
                             if time.time() - time_begin > 5
                             else ""
@@ -222,6 +264,7 @@ class GitAnnexHelper:
     @classmethod
     def subprocess(cls, fun, cmd, *args, **kwargs):
         logger.info(f"🚀 Running {cmd}")
+        cls.debug_cmdline(cmd)
         return fun(
             cmd, *args, **{**dict(encoding="utf-8", errors="ignore"), **kwargs}
         )
@@ -232,6 +275,8 @@ class GitAnnexHelper:
         cmdparts,
         cwd=None,
         terminal=True,
+        notify=False,
+        title=None,
         keep_open_on_success=False,
         keep_open_on_failure=True,
         dry_run=False,
@@ -273,24 +318,52 @@ class GitAnnexHelper:
                 f"|| ({close_cmd('💥 Failure!',timeout=None if keep_open_on_failure else 10)})",
             ]
             logger.debug(f"What the terminal will be given: {cmdparts = }")
-            if logger.getEffectiveLevel() <= logging.DEBUG:
-                console.print(f"📋 Copy-pastable:\n\n{shlex.join(cmdparts)}\n")
-            cmdparts = [
-                "xfce4-terminal",  # TODO: Hard-coded terminal emulator is bad
-                "--icon",
-                "git-annex",
-                "--hide-menubar",
-                "--hide-toolbar",
-                "--command",
-                shlex.join(cmdparts),
-            ]
-        logger.info(f"🚀 Running {cmdparts = }")
-        if logger.getEffectiveLevel() <= logging.DEBUG:
-            console.print(f"📋 Copy-pastable:\n\n{shlex.join(cmdparts)}\n")
+            cls.debug_cmdline(
+                cmdparts,
+                comment="What the Terminal will be given, 📋 copy-pastable:",
+            )
+            cmdparts = (
+                [
+                    "xfce4-terminal",  # TODO: Hard-coded terminal emulator is bad
+                    "--icon",
+                    "git-annex",
+                    "--hide-menubar",
+                    "--hide-toolbar",
+                ]
+                + (["--title", title] if title else [])
+                + [
+                    "--command",
+                    shlex.join(cmdparts),
+                ]
+            )
+        cls.debug_cmdline(
+            cmdparts,
+            comment="The entire command-line to run, 📋 copy-pastable:",
+        )
         if dry_run:
+            logger.info(f"🚀 Would now run {cmdparts = }")
             return cmdparts
         else:
-            return cls.subprocess(subprocess.run, cmdparts, cwd=cwd)
+            if notify:
+                notification = Notify.Notification.new(
+                    _("Git Annex Thunar Plugin"), None, "git-annex"
+                )
+                notification.update(
+                    _("Git Annex Thunar Plugin"),
+                    "⏳ "
+                    + (title or _("Running {cmd}...").format(cmd=cmdparts)),
+                    "git-annex",
+                )
+                notification.show()
+            result = cls.subprocess(subprocess.run, cmdparts, cwd=cwd)
+            if notify:
+                notification.update(
+                    _("Git Annex Thunar Plugin"),
+                    "✅ " + (title or _("Ran {cmd}").format(cmd=cmdparts)),
+                    "git-annex",
+                )
+                notification.show()
+            return result
 
     @classmethod
     def run_git_annex(
@@ -308,8 +381,7 @@ class GitAnnexHelper:
         paths = paths or []
         cmdparts = ["git", "annex", subcmd] + args + paths
         logger.debug(f"Bare {cmdparts = }")
-        if logger.getEffectiveLevel() <= logging.DEBUG:
-            console.print(f"📋 Copy-pastable:\n\n{shlex.join(cmdparts)}\n")
+        cls.debug_cmdline(cmdparts)
         if reset_before:
             cmdparts = [
                 "sh",
@@ -319,8 +391,7 @@ class GitAnnexHelper:
                 ),
             ]
             logger.debug(f"With git resetting: {cmdparts = }")
-            if logger.getEffectiveLevel() <= logging.DEBUG:
-                console.print(f"📋 Copy-pastable:\n\n{shlex.join(cmdparts)}\n")
+            cls.debug_cmdline(cmdparts)
         repo_description = cls.get_git_annex_description(cwd or ".")
         if commit_before:
             cmdparts = [
@@ -345,8 +416,7 @@ class GitAnnexHelper:
                 ),
             ]
             logger.debug(f"With git add: {cmdparts = }")
-            if logger.getEffectiveLevel() <= logging.DEBUG:
-                console.print(f"📋 Copy-pastable:\n\n{shlex.join(cmdparts)}\n")
+            cls.debug_cmdline(cmdparts)
         if add_before:
             cmdparts = [
                 "sh",
@@ -361,8 +431,7 @@ class GitAnnexHelper:
                 ),
             ]
             logger.debug(f"With git add: {cmdparts = }")
-            if logger.getEffectiveLevel() <= logging.DEBUG:
-                console.print(f"📋 Copy-pastable:\n\n{shlex.join(cmdparts)}\n")
+            cls.debug_cmdline(cmdparts)
         return cls.run_cmd(cmdparts=cmdparts, cwd=cwd, **kwargs)
 
     @classmethod
@@ -388,6 +457,7 @@ class GitAnnexHelper:
                         "annex",
                         "info",
                         "here",
+                        "--fast",
                         "--json",
                     ],
                 )
@@ -542,7 +612,22 @@ class GitAnnexSubmenu(GObject.GObject, Thunarx.MenuProvider, GitAnnexHelper):
             notify_flags = []
 
         def on_click(menuitem, func=self.run_git_annex, **kwargs):
-            kwargs = {**dict(cwd=cwd), **kwargs}
+            kwargs = {
+                **dict(
+                    cwd=cwd,
+                    title="{title} @ {repo}".format(
+                        title=kwargs.get(
+                            "title",
+                            shlex.join(
+                                ["git", "annex", kwargs.get("subcmd", "")]
+                                + kwargs.get("args", [])
+                            ),
+                        ),
+                        repo=pathlib.Path(repo).name,
+                    ),
+                ),
+                **kwargs,
+            }
             menuitem.connect(
                 "activate",
                 lambda item, *a, **kw: func(**kwargs),
@@ -594,8 +679,10 @@ class GitAnnexSubmenu(GObject.GObject, Thunarx.MenuProvider, GitAnnexHelper):
             on_click(
                 menuitem,
                 func=self.run_cmd,
-                cmdparts=["git", "switch", b],
+                cmdparts=(c := ["git", "switch", "--force", b]),
                 terminal=False,
+                notify=True,
+                title=f"{shlex.join(c)} @ {pathlib.Path(repo).name}",
             )
             git_switch_submenu.append_item(menuitem)
 
@@ -689,8 +776,10 @@ class GitAnnexSubmenu(GObject.GObject, Thunarx.MenuProvider, GitAnnexHelper):
         # cache rebuilding
         git_annex_metadata_rebuild_cache_menuitem = Thunarx.MenuItem(
             name="ContextMenu::GitAnnexMetadata::CacheRebuild",
-            label=_("Rebuild Cache"),
-            tooltip=_("Rebuild cache of Git Annex metadata"),
+            label=_("Build Metadata List"),
+            tooltip=_(
+                "Rebuild cache of Git Annex metadata to enable switching to views."
+            ),
             icon="database-index",
         )
         git_annex_metadata_submenu.append_item(
@@ -701,32 +790,101 @@ class GitAnnexSubmenu(GObject.GObject, Thunarx.MenuProvider, GitAnnexHelper):
             lambda item, *a, **kw: self.rebuild_metadata_cache(path=cwd, **kw),
         )
 
-        if len(METADATA_CACHE[uuid]["/"]) > 0 and not in_view:
-            git_annex_metadata_view_menuitem = Thunarx.MenuItem(
-                name="ContextMenu::GitAnnexMetadata::View",
-                label=_("View"),
-                tooltip=_("Switch to a metadata-driven view"),
-                icon="view-hidden",
+        if in_view:
+            menuitem = Thunarx.MenuItem(
+                name=f"ContextMenu::GitAnnexMetadata::Vcycle",
+                label=_("Cycle View"),
+                tooltip=_("Cycle the view layers"),
+                icon="object-rotate-left",
             )
+            git_annex_metadata_submenu.append_item(menuitem)
+            on_click(
+                menuitem,
+                subcmd="vcycle",
+                terminal=False,
+                notify=True,
+            )
+            menuitem = Thunarx.MenuItem(
+                name=f"ContextMenu::GitAnnexMetadata::Vpop",
+                label=_("Back to previous view"),
+                tooltip=_("Go back to last view"),
+                icon="edit-undo",
+            )
+            git_annex_metadata_submenu.append_item(menuitem)
+            on_click(
+                menuitem,
+                subcmd="vpop",
+                terminal=False,
+                notify=True,
+            )
+
+        git_annex_metadata_view_menuitem = Thunarx.MenuItem(
+            name="ContextMenu::GitAnnexMetadata::View",
+            label=_("Add to View") if in_view else _("View"),
+            tooltip=_("Modify current metadata-driven view")
+            if in_view
+            else _("Switch to a metadata-driven view"),
+            icon="view-hidden",
+        )
+        git_annex_metadata_view_submenu = Thunarx.Menu()
+        git_annex_metadata_view_menuitem.set_menu(
+            git_annex_metadata_view_submenu
+        )
+        menuitem = Thunarx.MenuItem(
+            name=f"ContextMenu::GitAnnexMetadata::View::**Folder**",
+            label=_("Path"),
+            tooltip=_(
+                "Add the file's folder structure to the current metadata-driven view"
+            )
+            if in_view
+            else _("Switch to a view that has the current folder structure."),
+            icon="folder",
+        )
+        git_annex_metadata_view_submenu.append_item(menuitem)
+        on_click(
+            menuitem,
+            subcmd=(c := "vadd" if in_view else "view"),
+            args=(a := [f"/=*"]),
+            terminal=False,
+            notify=True,
+        )
+        if len(METADATA_CACHE[uuid]["/"]) > 0 or in_view:
             git_annex_metadata_submenu.append_item(
                 git_annex_metadata_view_menuitem
             )
-            git_annex_metadata_view_submenu = Thunarx.Menu()
-            git_annex_metadata_view_menuitem.set_menu(
-                git_annex_metadata_view_submenu
-            )
+
+        if len(METADATA_CACHE[uuid]["/"]) > 0:
             for i, field in enumerate(METADATA_CACHE[uuid]["/"], start=1):
                 logger.debug(f"{i = }, {field = }")
                 menuitem = Thunarx.MenuItem(
                     name=f"ContextMenu::GitAnnexMetadata::View::{i}",
                     label=field,
-                    tooltip=_(
-                        "Switch to a metadata-driven view over field {field!r}"
-                    ).format(field=field),
+                    tooltip=(
+                        _(
+                            "Add a {field} level to the current metadata-driven view"
+                        ).format(field=field)
+                        if in_view
+                        else _(
+                            "Switch to a metadata-driven view over field {field}"
+                        )
+                    ).format(field=repr(field)),
                     icon="view-hidden",
                 )
                 git_annex_metadata_view_submenu.append_item(menuitem)
-                on_click(menuitem, subcmd="view", args=[f"{field}=*"])
+                on_click(
+                    menuitem,
+                    subcmd=(c := "vadd" if in_view else "view"),
+                    args=(
+                        a := [
+                            f"{field}?=*"
+                            if "view-show-valueless"
+                            in self.git_annex_capabilities()
+                            else f"{field}=*"
+                        ]
+                    ),
+                    terminal=False,
+                    notify=True,
+                )
 
         logger.debug(f"Assembling menus...")
         git_annex_submenu = Thunarx.Menu()
@@ -753,6 +911,26 @@ class GitAnnexProperties(
     def __init__(self):
         logger.debug(f"{type(self).__name__} is initialized")
 
+    @classmethod
+    def name(cls):
+        s = _("Git Annex Properties Page")
+        if not shutil.which("git-annex"):
+            s = _("[Unavailable]") + " " + s
+        return s
+
+    @classmethod
+    def description(cls):
+        s = _(
+            "This plugin adds a properties page "
+            "for managing Git Annex metadata."
+        )
+        if not shutil.which("git-annex"):
+            s += " " + _(
+                "Install Git Annex to use this plugin: "
+                "https://git-annex.branchable.com"
+            )
+        return s
+
     @property
     def ui(self):
         try:
